/* pam_server.c  - part of pam_watch.c
 * AUTHOR               : Samuel Behan <behan@frida.fri.utc.sk> (c) 2001
 * LICENSE      : GNU GPL v2 or later, NO WARANTY VOID provided
 * NOTE         : Modified by Samule Behan <behan@frida.fri.utc.sk> 
 * from original pam_filter by Andrew Morgan <morgan@transmeta.com> 
 * and (his) Richard Stevens' UNIX Network Programming book.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <termios.h>
#include <signal.h>
#include <pwd.h>
#include <time.h>
#include <errno.h>

#include <security/pam_modules.h>

#include "cscreen.h"
#include "pam_watch.h"

/* globals -- need to use them since they are used more times in 
 * situations when thay can not be passed as parameters */
static struct termios tty_mode;
static struct sigaction sact;
static watch_config_t *watch;
static cscreen_t *cscreen;

/*child IO streams */
#define APPIN_FILENO	3
#define APPOUT_FILENO	4
#define APPERR_FILENO	5


#define streq(str1, str2) (strcmp((str1), (str2)) == 0)
/* 'patches' current service if neccessary */
static int patch_service(pam_handle_t * pamh)
{
    /* for login service */
    char *service;

    pam_get_item(pamh, PAM_SERVICE, (void *) &service);
    alarm((unsigned int) 0);
    signal(SIGALRM, SIG_DFL);

    return 0;
}
#undef streq

static int propagate_session(pam_handle_t * pamh)
{
    int prop_f;
    static char *service_name = NULL;

    if (watch->propagate_name != NULL)
	free(watch->propagate_name);

    if (pamh != NULL)
	pam_get_item(pamh, PAM_SERVICE, (void *) &service_name);
    if (service_name == NULL)
	return PAM_ABORT;
    if ((watch->propagate_name = (char *) malloc(sizeof(char) *
						  (strlen(service_name) +
						   10))) == NULL)
    {
	ON_DEBUG(_pam_log
		 (LOG_CRIT, "can not malloc template propagate_name"));
	return PAM_ABORT;
    }
    sprintf(watch->propagate_name, "%s-%d+", service_name, getpid());
    /*security patch: file names are almost predictable - we rather
     * remove them, than killing some importatn file */
    unlink(watch->propagate_name);
    if ((prop_f =
	 open(watch->propagate_name, O_WRONLY | O_CREAT | O_NOFOLLOW,
	      0770)) == -1)
    {
	ON_DEBUG(_pam_log
		 (LOG_WARNING,
		  "can not create propagation file (%s), ignoring...",
		  watch->propagate_name));
	return PAM_ABORT;
    } else
    {
	char *tmp = (char *) malloc(sizeof(char) * (30));

	if (tmp == NULL)
	{
	    ON_DEBUG(_pam_log
		     (LOG_CRIT, "can not malloc for propagation file"));
	    return PAM_ABORT;
	}
	write(prop_f, "#!/bin/sh\n", 10);
	sprintf(tmp, "PID=%d\nSERVICE=", getpid());
	write(prop_f, tmp, strlen(tmp));
	write(prop_f, service_name, strlen(service_name));
	if (watch->suspended)
	    write(prop_f, "\nSUSPENDED=1\n", 13);
	write(prop_f,
	      "\nif [ -z \"$PAM_WATCHER\" ];\n\tthen PAM_WATCHER=`type -p pam_watcher`;\nfi\n",
	      72);
	write(prop_f,
	      "if [ -z \"$PAM_WATCHER\" ];\n\tthen echo 'Can not find pam_watcher script...';\n",
	      75);
	write(prop_f,
	      "\t      echo '  ...set $PAM_WATCHER to path to pam_watcher script (see README for more info)';\n",
	      94);
	write(prop_f, "\t      exit 1;\nfi\n\n", 19);
	write(prop_f, "exec $PAM_WATCHER $PID:$SERVICE:$SUSPENDED $@\n", 46);

	free(tmp);
    }

    close(prop_f);
    return PAM_SUCCESS;
}

static int create_output_fifo()
{
    if (watch->ofifo_name != NULL)
	free(watch->ofifo_name);

    if ((watch->ofifo_name = (char *) malloc(sizeof(char) *
					   (strlen(watch->propagate_name) +
					    1))) == NULL)
    {
	ON_DEBUG(_pam_log(LOG_CRIT, "can not malloc at ofifo_name"));
	return PAM_ABORT;
    }
    /* create link name */
    strcpy(watch->ofifo_name, watch->propagate_name);
    watch->ofifo_name[strlen(watch->ofifo_name) - 1] = '>';

    unlink(watch->ofifo_name);
    if (mkfifo(watch->ofifo_name, 0660) == -1)
    {
	ON_DEBUG(_pam_log(LOG_WARNING, "unable to create output fifo"));
	return PAM_ABORT;
    }
    return PAM_SUCCESS;
}

static int create_input_fifo()
{
    if (watch->ififo_name != NULL)
	free(watch->ififo_name);

    if (
	   (watch->ififo_name =
	    (char *) malloc(sizeof(char) *
			     (strlen(watch->propagate_name) + 1))) == NULL)
    {
	ON_DEBUG(_pam_log(LOG_CRIT, "can not malloc at ififo_name"));
	return PAM_ABORT;
    }
    /* create input fifo name from ofifo name */
    strcpy(watch->ififo_name, watch->propagate_name);
    watch->ififo_name[strlen(watch->ififo_name) - 1] = '<';

    unlink(watch->ififo_name);
    if (mkfifo(watch->ififo_name, 0660) == -1)
    {
	ON_DEBUG(_pam_log(LOG_WARNING, "unable to create input fifo"));
	return PAM_ABORT;
    }
    return PAM_SUCCESS;
}

static int open_output_fifo()
{
    watch->output_fifo = open(watch->ofifo_name,
			      O_WRONLY | O_NONBLOCK | O_NOFOLLOW);
    if (watch->output_fifo != -1)
    {
	ON_DEBUG(_pam_log(LOG_NOTICE, "watcher connected to output fifo"));
	return PAM_SUCCESS;
    }
    return PAM_ABORT;
}

static int open_input_fifo()
{
    if (watch->input_fifo != -1)
    {
	ON_DEBUG(_pam_log(LOG_NOTICE, "input fifo closed"));
	close(watch->input_fifo);
    }
    watch->input_fifo =
	open(watch->ififo_name, O_RDONLY | O_NONBLOCK | O_NOFOLLOW);
    if (watch->input_fifo != -1)
    {
	ON_DEBUG(_pam_log(LOG_NOTICE, "input fifo (re)opened"));
	return PAM_SUCCESS;
    }
    return PAM_ABORT;
}

#ifdef SUSPEND_STRING
static int suspend_session(pam_handle_t * pamh)
{
    void (*handler) (int);
    pid_t pid;

    /*NOTE: LOGIN patch - login is often the top session leader so quiting it 
     * will kill all its child by SIGHUP - so we ignore it while suspending
     * and then restore it's previous handler */
    handler = signal(SIGHUP, SIG_IGN);
    if ((pid = fork()) && (pid != -1))
    {				/* root exit() */
	fprintf(stderr, "\033[0G\033[1;0m[%d] Suspending session...\r\n", pid);
	ON_DEBUG(_pam_log(LOG_NOTICE, "suspending session"));
	exit(0);
    } else if (pid == -1)
    {
	fprintf(stderr, "\033[1;0mCan not suspend session (%s) !!!\r\n",
		strerror(errno));
	ON_DEBUG(_pam_log
		 (LOG_ERR, "failed to suspend session - fork() error"));
	return PAM_ABORT;
    }
    signal(SIGHUP, handler);

    watch->suspended = 1;
    /* propagate suspende session */
    if (propagate_session(pamh) != PAM_SUCCESS)
    {
	fprintf(stderr,
	    "\0333[1;0mFailed to propagate suspended session (pid %d)\r\n",
		getpid());
	ON_DEBUG(_pam_log(LOG_ERR, "failed to propagate suspended session"));
	return PAM_ABORT;
    }
    /* close IOs */
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
    dup2(APPIN_FILENO, STDIN_FILENO);
    dup2(APPOUT_FILENO, STDOUT_FILENO);
    dup2(APPERR_FILENO, STDERR_FILENO);
    fcntl(STDIN_FILENO, F_SETFD, 0);
    fcntl(STDOUT_FILENO, F_SETFD, 0);
    fcntl(STDERR_FILENO, F_SETFD, 0);
    /* get our control terminal, back */
    if (ioctl(APPIN_FILENO, TIOCSCTTY, APPIN_FILENO) == -1)
    {
	ON_DEVEL(_pam_log(LOG_WARNING,
			  "child can not set controling terminal - %s",
			  strerror(errno)));
    }
    /* setpgid */
/*
 * FIXME: I dont know how to get back to process group, It is not very needed
 * but without this we will miss some important events that can occur in the 
 * system.
 */
#ifdef _TRY_SETGID
    if (setpgid(watch->child_pid, pid) == -1)
    {
	ON_DEVEL(_pam_log(LOG_DEBUG, "could not setpgid(%d, %d) - %s", watch->child_pid,
			  pid, strerror(errno)));
    }
#endif
    return PAM_SUCCESS;
}
#endif				/*SUSPEND_STRING */

/* 
 * SIGNAL COMMUNICATION MODEL ( used because it si very, very simple )
 *   -- I use unix signals as it were 'bits' and from its position in set
 * it is created specific integer number, unique for any signal set
 *   -- for sequence of 3 signals we can get 8 different commands
 * mathematicaly:
 * SIGUSR1 = 0:  command_id |= 01*pow(2, signal_position);
 * SIGUSR2 = 1:  command_id |= 04*pow(2, signal_position);
 * NOTE: Each non complete signal set expires after 10 seconds from the first
 * received signal
 */

#define CMDEXP 10		/*signal set expiration time */
#define CMDSET 3		/*length of of one signal set */
#define CMD111 56
#define CMD110 24
#define CMD100 12
#define CMD000 14
#define CMD001 38
#define CMD011 50
#define CMD010 26
#define CMD101 44

/* executes given command */
static int exec_command(int cmd_num)
{
    switch (cmd_num)
    {
	case CMD111:		/* io copying required -- create/open fifos, print */
	    {			/* current screen and start io copying */
		ON_DEVEL(_pam_log(LOG_DEBUG, "sigcmd: io open required"));
		if (watch->output_fifo != -1)
		{
		    ON_DEBUG(_pam_log
			     (LOG_WARNING,
			      "can not open IO, already watching"));
		}
		if (create_input_fifo() != PAM_SUCCESS)
		{
		    ON_DEBUG(_pam_log(LOG_ERR, "can not create input fifo"));
		    return PAM_ABORT;
		}
		if (create_output_fifo() != PAM_SUCCESS)
		{
		    ON_DEBUG(_pam_log
			     (LOG_ERR, "can not create output fifo"));
		    return PAM_ABORT;
		}
		if (open_input_fifo() != PAM_SUCCESS)
		{
		    ON_DEBUG(_pam_log(LOG_ERR, "can not open input fifo"));
		    return PAM_ABORT;
		}
		watch->require_io = 1;	/* initiate conection accept */
	    }
	    break;
	case CMD110:		/* print current user screen */
	    {
		ON_DEVEL(_pam_log
			 (LOG_DEBUG, "sigcmd: print current console"));
		if (watch->output_fifo != -1 && watch->cache_enabled)
		    cscreen_print(cscreen, watch->output_fifo);
	    }
	    break;
	case CMD100:		/* toggle user console input */
	    {
		ON_DEVEL(_pam_log(LOG_DEBUG, "sigcmd: togle user input"));
		watch->disable_input = ((watch->disable_input) ? 0 : 1);
	    }
	    break;
	case CMD000:
	    {
		propagate_session(NULL);
	    }
	    break;
	case CMD001:		/* (re)echo user input */
	    {
		ON_DEVEL(_pam_log
			 (LOG_DEBUG, "sigcmd: togle user input echoing"));
		watch->echo_input = ((watch->echo_input) ? 0 : 1);
	    }
	    break;
	ON_DEVEL(default:
_pam_log(LOG_DEBUG, "got not implemented sigcommand"));
    }
    return PAM_SUCCESS;
}

#define CALLER_ONEXIT	1
#define CALLER_ERROR	2
#define CALLER_USER	3
#define CALLER_SUSPEND  4

static void exit_cleanup(int exit_code, void *caller)
{
    if (tcsetattr(STDIN_FILENO, TCSANOW, &tty_mode) == -1)
    {
	ON_DEBUG(_pam_log(LOG_WARNING,
			  "can not set back terminal mode on exit"));
    }
    /* remove fifos - they have been almost removed but.... */
    if (watch->ofifo_name != NULL)
	unlink(watch->ofifo_name);
    if (watch->ififo_name != NULL)
	unlink(watch->ififo_name);
    if (watch->propagate_name != NULL)
	unlink(watch->propagate_name);
}

/* signal catching handler */
void catch_signal(int signum)
{
    static time_t cmd_time = 0;
    static short int cmd_num = 0;
    static int cmd = 0;

    /* small (non recursive) integer implementation of pow() function */
    static int pow(int x, int y)
    {
	int res = 1;
	for (; y > 0; y--)
	     res *= x;
	 return res;
    }
    switch (signum)
    {
	case SIGPIPE:
	    {
		if (watch->output_fifo == -1)
		    break;
		if (close(watch->output_fifo) != -1)
		{
		    unlink(watch->ofifo_name);
		    watch->output_fifo = -1;
		    ON_DEBUG(_pam_log
			     (LOG_WARNING,
			      "watcher disconnected from output fifo"));
		} else
		    ON_DEBUG(_pam_log(LOG_ERR, "can not close output fifo"));
	    }
	    break;
	case SIGUSR1:
	case SIGUSR2:
	    {
		if (cmd_time == 0)
		    cmd_time = time(NULL);	/* initialize timer */
		else if ((time(NULL) - cmd_time) > CMDEXP)
		{
		    cmd_time = time(NULL);
		    cmd_num = cmd = 0;
		}
		cmd_num++;
		cmd |= ((signum == SIGUSR1) ? 01 : 04) * pow(2, cmd_num);
		if (cmd_num == CMDSET)
		{		/* signal set complete */
		    ON_DEVEL(_pam_log
			     (LOG_DEBUG, "executing sigcommand number '%d'",
			      cmd));
		    exec_command(cmd);
		    cmd_num = cmd = cmd_time = 0;
		}
	    }
	    break;
	case SIGCHLD:
	    {
		int state;

		ON_DEVEL(_pam_log(LOG_DEBUG, "exit() on sigchld"));
		wait(&state);
		exit(state);
	    }
	    break;
#ifdef SIGWINCH			/*since it is not a POSIX signal */
	case SIGWINCH:
	    {
		struct winsize win =
		{0, 0, 0, 0};

		if (watch->suspended)
		    break;	/*suspended can not */
		ON_DEVEL(_pam_log(LOG_DEBUG, "got SIGWINCH - resending"));
		if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &win) == -1
		    || win.ws_col == 0 || win.ws_row == 0)
		    ON_DEVEL(_pam_log(LOG_DEBUG, "failed to get new winsize"));
		else if (ioctl(APPOUT_FILENO, TIOCSWINSZ, &win) == -1)
		    ON_DEVEL(_pam_log(LOG_DEBUG, "failed to set new winsize"));
		
	    }
	    break;
#endif

	case SIGSEGV:
	    {
		write(STDOUT_FILENO, "\r\n\033[1;31mGOT SIGSEGV - report error\033[1;0m\r\n", 27);
		exit(errno);
	    }
	    break;
	default:
	    ON_DEVEL(_pam_log(LOG_DEBUG, "non handled signal received"));
    }
    /* reentrancy ??? */
    sigaction(signum, &sact, NULL);
}


/*------------------------------------------------------------------------------
********************************************************************************
--------------------------------------------------------------------------------*/
int start_server(pam_handle_t * pamh, int fd[2], struct termios old_mode,
		 register watch_config_t * watch_ref)
{
    fd_set readers;
    char *user_name;
    char buffer[BUFSIZ];

    /* globalize - some function can not receive parameters, so we need this */
    tty_mode = old_mode;
    watch = watch_ref;

    /*try to drop privileges */
    pam_get_item(pamh, PAM_USER, (void *) &user_name);

    if (user_name == NULL)
    {
	_pam_log(LOG_CRIT, "can not get user name (broken pam ???)");
	exit(errno);
    }
#ifdef DROP_PRIVILEGES
    else
    {
	struct passwd *pw = NULL;

	setpwent();
	if ((pw = getpwnam(user_name)) != NULL)
	{
	    setreuid(pw->pw_uid, pw->pw_uid);
	    setregid(pw->pw_gid, pw->pw_gid);
	}
	ON_DEVEL(
		    else
		    _pam_log(LOG_DEBUG, "can not retreive user uid"));
	endpwent();
    }
#endif

    /* register on exit fucntion */
    if (on_exit(*exit_cleanup, (void *) CALLER_ONEXIT) == -1)
    {
	ON_DEBUG(_pam_log(LOG_WARNING, "can not register at exit function"));
	exit_cleanup(0, (void *) CALLER_ERROR);
	exit(0);
    }
    /* initialize working dir */
    if ((chdir(watch->path) == -1)
	|| (mkdir(user_name, 0770) == -1 && errno != EEXIST)
	|| (chdir(user_name) == -1))
    {
	ON_DEBUG(_pam_log(LOG_ERR, "can not init work dir"));
	exit(errno);
    }
    /* setup master side interlink */
    if (dup2(fd[0], APPIN_FILENO) == -1
	|| dup2(fd[0], APPOUT_FILENO) == -1
	|| dup2(fd[0], APPERR_FILENO) == -1)
    {
	ON_DEBUG(_pam_log(LOG_ERR,
			  "can not assign controling STDIN/OUT/ERR"));
	exit(errno);
    }
    if ((fcntl(APPIN_FILENO, F_SETFD, 0)) == -1
	|| (fcntl(APPOUT_FILENO, F_SETFD, 0)) == -1
	|| (fcntl(APPERR_FILENO, F_SETFD, 0)) == -1)
    {
	ON_DEBUG(_pam_log
		 (LOG_WARNING, "unable to retain controling I/O modes"));
	exit(errno);
    }
    /*initializa screen cacher */
    if (watch->cache_enabled)
	if ((cscreen = cscreen_open(watch->cache_size)) == NULL)
	{
	    ON_DEBUG(_pam_log
		     (LOG_WARNING, "error initiliazing screen cacher"));
	    watch->cache_enabled = 0;
	}
    /* initialize and propagate session */
    if (propagate_session(pamh) != PAM_SUCCESS)
    {
	ON_DEVEL(_pam_log
		 (LOG_DEBUG, "failed to create propagation file (%s)",
		  strerror(errno)));
	exit(errno);
    }
    /* patch current service functions */
    patch_service(pamh);

#ifdef __DESTROY_PAM_HANDLER
    /* dealloc everythink (look for pam) */
    (void) pam_set_item(pamh, PAM_AUTHTOK, NULL);
    (void) pam_set_item(pamh, PAM_OLDAUTHTOK, NULL);
#endif

    /* install signal handler - using BSD's SA_RESTART to not break
     * IO primitives by the signals */
    sact.sa_handler = *catch_signal;	/*handler */
    sact.sa_flags = SA_RESTART;	/*BSD's restarter */
    sigemptyset(&sact.sa_mask);	/*block for SIGCMDs */
    sigaddset(&sact.sa_mask, SIGUSR1);
    sigaddset(&sact.sa_mask, SIGUSR2);
    sigaction(SIGUSR1, &sact, NULL);	/*install handlers for... */
    sigaction(SIGUSR2, &sact, NULL);	/* .. SIGCMD's */
    sigaction(SIGPIPE, &sact, NULL);	/* watcher disconnect */
    sigaction(SIGCHLD, &sact, NULL);	/* child end */
#ifdef SIGWINCH
    sigaction(SIGWINCH, &sact, NULL);	/*session win resize */
#endif
    ON_DEVEL(sigaction(SIGSEGV, &sact, NULL));	/* developers */

/*
 * FIXME: Add select check while writing to watcher
 */

    /* start I/O copying & watching */
    while (1)
    {
	FD_ZERO(&readers);

	/* open output fifo if reader avaible (only on SIGUSR1) */
	if ((watch->output_fifo == -1) && (watch->require_io))
	{
	    if (open_output_fifo() == PAM_SUCCESS)
		watch->require_io = 0;
	}
	/* dancing round the console I/O */
	FD_SET(APPOUT_FILENO, &readers);
	FD_SET(APPERR_FILENO, &readers);
	/* user console input */
	if (!watch->suspended)
	    FD_SET(STDIN_FILENO, &readers);
	/* input fifo */
	if (!(watch->ctrl & WATCH_NOWRITE) && (watch->input_fifo != -1))
	    FD_SET(watch->input_fifo, &readers);

	if (select(APPERR_FILENO + 2, &readers, NULL, NULL, NULL) < 0)
	{
	    if (errno != EINTR)
	    {
		ON_DEBUG(_pam_log(LOG_WARNING, "select failed (%s)",
				  strerror(errno)));
		break;
	    } else		/* signal has been received */
		continue;
	}
	/* user output pipe */
	if (FD_ISSET(APPOUT_FILENO, &readers))
	{
	    int got = read(APPOUT_FILENO, buffer, BUFSIZ);

#ifdef SUSPEND_STRING
	    if (!(watch->ctrl & WATCH_NOSUSPEND)
		&& !watch->suspended
		&& got == 3
		&& buffer[0] == SUSPEND_STRING[0]
		&& buffer[1] == SUSPEND_STRING[1]
		&& buffer[2] == SUSPEND_STRING[2])
	    {
		if (suspend_session(pamh) != PAM_SUCCESS)
		    ON_DEBUG(_pam_log(LOG_ERR,
			     "can not suspend terminal on user request "));
	    } else
#endif
	    if (got != -1)
	    {
		if (!watch->suspended)
		    write(STDOUT_FILENO, buffer, got);
		if (watch->cache_enabled)
		    if (cscreen_write(cscreen, buffer, got) == -1)
		    {
			ON_DEBUG(_pam_log(LOG_ERR,
				       "error caching screen, disabling"));
			watch->cache_enabled = 0;
		    }
		if (watch->output_fifo != -1)
		    write(watch->output_fifo, buffer, got);
	    } else
		break;
	}
	/* user input pipe */
	if (FD_ISSET(STDIN_FILENO, &readers) && !watch->suspended)
	{
	    int got = read(STDIN_FILENO, buffer, BUFSIZ);

	    if (got != -1)
	    {
		if (!watch->disable_input)	/* disable user input */
		    write(APPIN_FILENO, buffer, got);
		if (watch->echo_input)	/* echo user input */
		    write(watch->output_fifo, buffer, got);
	    } else
		break;
	}
	/* watcher input pipe */
	if (!(watch->ctrl & WATCH_NOWRITE)
	    && (watch->input_fifo != -1)
	    && FD_ISSET(watch->input_fifo, &readers))
	{
	    int got = read(watch->input_fifo, buffer, BUFSIZ);

	    if (got == 0)
	    {
		close(watch->input_fifo);
		unlink(watch->ififo_name);
		watch->input_fifo = -1;
	    }
#ifdef SUSPEND_RESIZE
	    else if (got <= 10
		     && got > 3
		     && buffer[0] == SUSPEND_RESIZE[0]
		     && buffer[1] == SUSPEND_RESIZE[1]
		     && buffer[2] == SUSPEND_RESIZE[2])
	    {			/* parse string in format INTxINT */
		short int i = 0, ii = 0;
		char snum[4] = "\0\0\0\0";
		struct winsize win =
		{0, 0, 0, 0};

		for (i = 0; (i < 3 && buffer[3 + i] != 'x' && i < got - 3);
		     snum[i] = buffer[3 + i++]);
		win.ws_row = abs(atoi(snum));	/*convert char->int */
		snum[0] = snum[1] = snum[2] = snum[3] = '\0';
		i++;		/*skip 'x' */
		for (ii = 0; (ii < 3 && ii < got - 3 - i);
		     snum[ii] = buffer[3 + i + ii++]);
		win.ws_col = abs(atoi(snum));	/*convert char->int */
		ON_DEVEL(_pam_log(LOG_DEBUG, "remote win resize : %dx%d", win.ws_row, win.ws_col));
		if (win.ws_row != 0 && win.ws_col != 0)
		{
		    if (ioctl(APPIN_FILENO, TIOCSWINSZ, &win) == -1)
			ON_DEVEL(_pam_log(LOG_DEBUG, "can not set remote-window size"));
		} else
		    ON_DEVEL(_pam_log(LOG_DEBUG, "remote win_resize parse error %dx%d", win.ws_row, win.ws_col));
	    }
#endif
	    else if (got != -1)
	    {
		write(APPIN_FILENO, buffer, got);
	    } else
		break;
	}
    }

    /* we got eof */
    exit(0);
    return PAM_ABORT;
}

 /* EOF */
